@@ -10,6 +10,7 @@ module Network.HTTP.ReverseProxy
       ProxyDest (..)
       -- * Raw
     , rawProxyTo
+    , rawTcpProxyTo
       -- * WAI + http-conduit
     , waiProxyTo
     , defaultOnExc
@@ -58,13 +59,15 @@ import qualified Data.Conduit.Network           as DCN
 import           Data.Default.Class             (Default (..), def)
 import           Data.Functor.Identity          (Identity (..))
 import           Data.IORef
-import           Data.Maybe                     (fromMaybe)
+import           Data.Maybe                     (fromMaybe, listToMaybe)
 import           Data.Monoid                    (mappend, mconcat, (<>))
 import           Data.Set                       (Set)
 import qualified Data.Set                       as Set
 import           Data.Streaming.Network         (AppData, readLens)
 import qualified Data.Text.Lazy                 as TL
 import qualified Data.Text.Lazy.Encoding        as TLE
+import qualified Data.Text                      as T
+import qualified Data.Text.Encoding             as TE
 import           Data.Word8                     (isSpace, _colon, _cr)
 import           Network.HTTP.Client            (BodyReader, brRead)
 import qualified Network.HTTP.Client            as HC
@@ -119,16 +122,36 @@ rawProxyTo getDest appdata = do
   where
     fromClient = DCN.appSource appdata
     toClient = DCN.appSink appdata
-    withServer rsrc appdataServer = do
-        x <- newEmptyMVar
-        tid1 <- fork $ (rsrc $$+- toServer) `finally` putMVar x True
-        tid2 <- fork $ (fromServer $$ toClient) `finally` putMVar x False
-        y <- takeMVar x
-        killThread $ if y then tid2 else tid1
+    withServer rsrc appdataServer = void $ concurrently
+        (rsrc $$+- toServer)
+        (fromServer $$ toClient)
       where
         fromServer = DCN.appSource appdataServer
         toServer = DCN.appSink appdataServer
 
+-- | Set up a reverse tcp proxy server, which will have a minimal overhead.
+--
+-- This function uses raw sockets, parsing as little of the request as
+-- possible. The workflow is:
+--
+-- 1. Open up a connection to the given host\/port.
+--
+-- 2. Pass all bytes across the wire unchanged.
+--
+-- If you need more control, such as modifying the request or response, use 'waiProxyTo'.
+--
+-- Since 0.4.4
+rawTcpProxyTo :: (MonadBaseControl IO m, MonadIO m)
+           => ProxyDest
+           -> AppData
+           -> m ()
+rawTcpProxyTo (ProxyDest host port) appdata = liftIO $
+    DCN.runTCPClient (DCN.clientSettings port host) withServer
+  where
+    withServer appdataServer = void $ concurrently
+      (DCN.appSource appdata       $$ DCN.appSink appdataServer)
+      (DCN.appSource appdataServer $$ DCN.appSink appdata      )
+
 -- | Sends a simple 502 bad gateway error message with the contents of the
 -- exception.
 defaultOnExc :: SomeException -> WAI.Application
@@ -319,11 +342,14 @@ fixReqHeaders wps req =
                $ WAI.requestHeaders req
   where
     fromSocket = (("X-Real-IP", S8.pack $ showSockAddr $ WAI.remoteHost req):)
+    fromForwardedFor = do
+      h <- lookup "x-forwarded-for" (WAI.requestHeaders req)
+      listToMaybe $ map (TE.encodeUtf8 . T.strip) $ T.splitOn "," $ TE.decodeUtf8 h
     addXRealIP =
         case wpsSetIpHeader wps of
             SIHFromSocket -> fromSocket
             SIHFromHeader ->
-                case lookup "x-real-ip" (WAI.requestHeaders req) <|> lookup "X-Forwarded-For" (WAI.requestHeaders req) of
+                case lookup "x-real-ip" (WAI.requestHeaders req) <|> fromForwardedFor of
                     Nothing -> fromSocket
                     Just ip -> (("X-Real-IP", ip):)
             SIHNone -> id
